"""Module defining all data primitives used in the protocol messages

Field metadata:

During (de)serialization the `metadata` parameter of the `dataclasses.field`
function to control how to perform (de)serialization. See
`aioslsk.protocol.primitives` for how (de)serialization is performed

These metadata keys are implemented:

* 'type': <type_class>
** defines the primary type of the data

* 'subtype': <type_class>
** used for arrays : the type of the elements contained in the array

* 'if_true': <field_name>
** serialization : only pack this field if the value of field with name <field_name> evaluates to True
** deserialization : only parse this field if the value of field with name <field_name> evaluates to True

* 'if_false': <field_name>
** serialization : only pack this field if the value of field with name <field_name> evaluates to False
** deserialization : only parse this field if the value of field with name <field_name> evaluates to False

* 'optional': True
** serialization : only pack this field if its value is anything other than None
** deserialization : during deserialization the code will determine if the message
    has been fully parsed. If not it will parse this field
"""
from dataclasses import dataclass, field, Field, fields, is_dataclass
import hashlib
import logging
import socket
import struct
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Optional,
    Protocol,
    Tuple,
    Type,
    TypeVar,
    Union
)
import zlib

logger = logging.getLogger(__name__)


T = TypeVar('T', bound='Serializable')

class Serializable(Protocol[T]):

    @classmethod
    def deserialize(cls, pos: int, data: bytes) -> Tuple[int, T]:
        ...

    def serialize(self) -> bytes:
        ...


def decode_string(value: bytes) -> str:
    try:
        return value.decode('utf-8')
    except UnicodeDecodeError:
        return value.decode('cp1252')


class uint8(int):
    STRUCT = struct.Struct('<B')

    def serialize(self) -> bytes:
        return self.STRUCT.pack(self)

    @classmethod
    def deserialize(cls, pos: int, data: bytes) -> Tuple[int, int]:
        return pos + cls.STRUCT.size, cls.STRUCT.unpack_from(data, offset=pos)[0]


class uint16(int):
    STRUCT = struct.Struct('<H')

    def serialize(self) -> bytes:
        return self.STRUCT.pack(self)

    @classmethod
    def deserialize(cls, pos: int, data: bytes) -> Tuple[int, int]:
        return pos + cls.STRUCT.size, cls.STRUCT.unpack_from(data, offset=pos)[0]


class uint32(int):
    STRUCT = struct.Struct('<I')

    def serialize(self) -> bytes:
        return self.STRUCT.pack(self)

    @classmethod
    def deserialize(cls, pos: int, data: bytes) -> Tuple[int, int]:
        return pos + cls.STRUCT.size, cls.STRUCT.unpack_from(data, offset=pos)[0]


class uint64(int):
    STRUCT = struct.Struct('<Q')

    def serialize(self) -> bytes:
        return self.STRUCT.pack(self)

    @classmethod
    def deserialize(cls, pos: int, data: bytes) -> Tuple[int, int]:
        return pos + cls.STRUCT.size, cls.STRUCT.unpack_from(data, offset=pos)[0]


class int32(int):
    STRUCT = struct.Struct('<i')

    def serialize(self) -> bytes:
        return self.STRUCT.pack(self)

    @classmethod
    def deserialize(cls, pos: int, data: bytes) -> Tuple[int, int]:
        return pos + cls.STRUCT.size, cls.STRUCT.unpack_from(data, offset=pos)[0]


class string(str):

    def serialize(self, encoding: str = 'utf-8') -> bytes:
        byte_string = self.encode(encoding)

        return uint32(len(byte_string)).serialize() + byte_string

    @classmethod
    def deserialize(cls, pos: int, data: bytes) -> Tuple[int, str]:
        pos_after_len, length = uint32.deserialize(pos, data)
        value = data[pos_after_len:pos_after_len + length]
        if len(value) != length:
            raise Exception(
                f"expected string with length ({length}), got {len(value)}")
        return pos_after_len + length, decode_string(value)


class bytearr(bytes):

    def serialize(self) -> bytes:
        length = len(self)
        return uint32(length).serialize() + bytes(self)

    @classmethod
    def deserialize(cls, pos: int, data: bytes) -> Tuple[int, str]:
        pos_after_len, length = uint32.deserialize(pos, data)
        value = data[pos_after_len:pos_after_len + length]
        return pos_after_len + length, value


class ipaddr(str):
    STRUCT = struct.Struct('<4s')

    def serialize(self) -> bytes:
        ip_b = socket.inet_aton(self)
        return self.STRUCT.pack(bytes(reversed(ip_b)))

    @classmethod
    def deserialize(cls, pos: int, data) -> Tuple[int, str]:
        value = cls.STRUCT.unpack(data[pos:pos + 4])[0]
        ip_addr = socket.inet_ntoa(bytes(reversed(value)))
        return pos + 4, ip_addr


class boolean(int):
    STRUCT = struct.Struct('<?')

    def serialize(self) -> bytes:
        return self.STRUCT.pack(self)

    @classmethod
    def deserialize(cls, pos: int, data: bytes) -> Tuple[int, bool]:
        return pos + cls.STRUCT.size, cls.STRUCT.unpack_from(data, offset=pos)[0]


class array(list):

    def serialize(self, element_type: Type[T]) -> bytes:
        body = uint32(len(self)).serialize()
        is_protocoldc = is_dataclass(element_type)
        for value in self:
            if is_protocoldc:
                body += value.serialize()
            else:
                body += element_type(value).serialize()
        return body

    @classmethod
    def deserialize(cls, pos: int, data: bytes, element_type: Type[T]) -> Tuple[int, List[T]]:
        items = []
        pos, array_len = uint32.deserialize(pos, data)
        for _ in range(array_len):
            pos, item = element_type.deserialize(pos, data)
            items.append(item)
        return pos, items


class ProtocolDataclass:
    """The `ProtocolDataclass` defines a collection of primitives that can be
    serialized or deserialized. Classes inheriting from this class should use
    the `@dataclass(order=True)` decorator. The order needs to be kept as the
    fields definitions will be evaluated during (de)serialization.

    Example definition:

    @dataclass(order=True)
    class CustomDataclass(ProtocolDataclass):
        username: str = field(metadata={'type': string})
        password: str = field(metadata={'type': string})
        has_privileges: bool = field(metadata={'type': boolean})
    """
    _CACHED_FIELDS: Optional[Tuple[Field]] = None
    """Cache version of the `dataclasses.fields` return for the current class
    """

    def serialize(self) -> bytes:
        message = bytes()
        if self.__class__._CACHED_FIELDS is None:
            self.__class__._CACHED_FIELDS = fields(self)

        for obj_field in fields(self):
            value = self._get_value_for_field(self, obj_field)
            if value is None:
                continue

            # Serialize
            try:
                proto_type: Type[Serializable] = obj_field.metadata['type']
            except KeyError:
                raise Exception(f"no 'type' for field {obj_field.name!r} defined")

            if is_dataclass(proto_type):
                message += value.serialize()
            elif 'subtype' in obj_field.metadata:
                if is_dataclass(obj_field.metadata['subtype']):
                    # during serialization of elements in the array the code
                    # will try to wrap the values into the passed type (in
                    # order to call 'serialize' on them) but dataclasses don't
                    # need to be wrapped. Just use a dummy lambda that returns
                    # the dataclass object as-is
                    message += proto_type(value).serialize(lambda val: val)
                else:
                    message += proto_type(value).serialize(obj_field.metadata['subtype'])
            else:
                message += proto_type(value).serialize()

        return message

    @classmethod
    def deserialize(cls, pos: int, message: bytes):
        if cls._CACHED_FIELDS is None:
            cls._CACHED_FIELDS = fields(cls)

        field_map: Dict[str, Any] = {}
        for obj_field in cls._CACHED_FIELDS:
            if not cls._field_needs_deserialization(obj_field, field_map, pos, message):
                continue

            try:
                proto_type = obj_field.metadata['type']
            except KeyError:
                raise Exception(f"no 'type' for field {obj_field.name!r} defined")

            if isinstance(proto_type, ProtocolDataclass):
                pos, value = proto_type.deserialize(pos, message)
            elif 'subtype' in obj_field.metadata:
                pos, value = proto_type.deserialize(pos, message, obj_field.metadata['subtype'])
            else:
                pos, value = proto_type.deserialize(pos, message)

            field_map[obj_field.name] = value

        return pos, cls(**field_map)

    @classmethod
    def _field_needs_deserialization(cls, field: Field, field_map: Dict[str, Any], pos: int, message: bytes):
        # For if_true and if_false we need to return only if the condition is
        # is false as we still want to check the 'optional' field
        if 'if_true' in field.metadata:
            if not field_map[field.metadata['if_true']]:
                return False

        if 'if_false' in field.metadata:
            if field_map[field.metadata['if_false']]:
                return False

        if 'optional' in field.metadata:
            return has_unparsed_bytes(pos, message)

        return True

    def _get_value_for_field(self, obj, fld: Field) -> Optional[Serializable]:
        value = getattr(obj, fld.name)

        # Order of checking the metadata is important
        if 'optional' in fld.metadata:
            if value is None:
                return None

        if 'if_true' in fld.metadata:
            other_value = getattr(obj, fld.metadata['if_true'])
            return value if bool(other_value) else None

        if 'if_false' in fld.metadata:
            other_value = getattr(obj, fld.metadata['if_false'])
            return value if not bool(other_value) else None

        return value


class MessageDataclass(ProtocolDataclass):
    """Message data class for which protocol messages should inherit from. This
    takes all behaviour from the `ProtocolDataclass` class but adds:

    * Prepending the message with length and MESSAGE_ID
    * Optionally the message data will (de)compressed
    """
    MESSAGE_ID: ClassVar[Union[uint8, uint32]] = uint32(0x00)

    def serialize(self, compress: bool = False) -> bytes:
        """Serializes the current `MessageDataClass` object and prepends the
        message length and `MESSAGE_ID`

        In case the message needs to be compressed just override this method
        in the subclass and simply call the super method with `compress=True`

        :param compress: use gzip compression on the message contents
        """
        message = super().serialize()

        if compress:
            message_len_before = len(message)
            message = zlib.compress(message)
            logger.debug(f"compressed {message_len_before} to {len(message)} bytes")

        message = self.MESSAGE_ID.serialize() + message
        return uint32(len(message)).serialize() + message

    @classmethod
    def deserialize(cls, message: bytes, decompress: bool = False):
        """Deserializes the passed `message` into an object of the current type

        In case the message needs to be decompressed just override this method
        in the subclass and simply call the super method with `decompress=True`

        :param decompress: use gzip decompression on the message contents
        :raise ValueError: if the message_id found the data does not match the
            `MESSAGE_ID` defined in the current class
        :return: an object of the current class
        """
        pos: int = 0

        # Parse length and header
        pos, _ = uint32.deserialize(pos, message)
        pos, message_id = type(cls.MESSAGE_ID).deserialize(pos, message)
        if message_id != cls.MESSAGE_ID:
            raise ValueError(f"message id mismatch {message_id} != {cls.MESSAGE_ID}")

        if decompress:
            message_len_before = len(message)
            message = zlib.decompress(message[pos:])
            logger.debug(f"decompressed {message_len_before} to {len(message)} bytes")

            pos, obj = super().deserialize(0, message)
        else:
            pos, obj = super().deserialize(pos, message)

        if has_unparsed_bytes(pos, message):
            logger.warning(f"message has {len(message[pos:])} unparsed bytes : {message!r}")

        return obj


_ATTR_STRUCT = struct.Struct('<II')

@dataclass(frozen=True, order=True)
class Attribute(ProtocolDataclass):
    key: int = field(metadata={'type': uint32})
    value: int = field(metadata={'type': uint32})

    @classmethod
    def deserialize(cls, pos: int, message: bytes):
        return (
            pos + _ATTR_STRUCT.size,
            cls(*_ATTR_STRUCT.unpack_from(message, offset=pos))
        )

    def serialize(self) -> bytes:
        return _ATTR_STRUCT.pack(self.key, self.value)


@dataclass(frozen=True, order=True)
class SimilarUser(ProtocolDataclass):
    username: str = field(metadata={'type': string})
    status: int = field(metadata={'type': uint32})


@dataclass(frozen=True, order=True)
class Recommendation(ProtocolDataclass):
    recommendation: str = field(metadata={'type': string})
    score: int = field(metadata={'type': int32})


@dataclass(frozen=True, order=True)
class RoomTicker(ProtocolDataclass):
    username: str = field(metadata={'type': string})
    ticker: str = field(metadata={'type': string})


@dataclass(frozen=True, order=True)
class PotentialParent(ProtocolDataclass):
    username: str = field(metadata={'type': string})
    ip: str = field(metadata={'type': ipaddr})
    port: int = field(metadata={'type': uint32})


@dataclass(frozen=True, order=True)
class UserStats(ProtocolDataclass):
    avg_speed: int = field(metadata={'type': uint32})
    uploads: int = field(metadata={'type': uint64})
    shared_file_count: int = field(metadata={'type': uint32})
    shared_folder_count: int = field(metadata={'type': uint32})


@dataclass(frozen=True, order=True)
class FileData(ProtocolDataclass):
    unknown: int = field(metadata={'type': uint8})
    filename: str = field(metadata={'type': string})
    filesize: int = field(metadata={'type': uint64})
    extension: str = field(metadata={'type': string})
    attributes: List[Attribute] = field(metadata={'type': array, 'subtype': Attribute})

    @classmethod
    def deserialize(cls, pos: int, message: bytes):
        pos, unknown = uint8.deserialize(pos, message)
        pos, filename = string.deserialize(pos, message)
        pos, filesize = uint64.deserialize(pos, message)
        pos, ext = string.deserialize(pos, message)
        pos, attrs = array.deserialize(pos, message, Attribute)
        return pos, cls(
            unknown=unknown,
            filename=filename,
            filesize=filesize,
            extension=ext,
            attributes=attrs
        )

    def serialize(self) -> bytes:
        return (
            uint8(self.unknown).serialize() +
            string(self.filename).serialize() +
            uint64(self.filesize).serialize() +
            string(self.extension).serialize() +
            array(self.attributes).serialize(Attribute)
        )


@dataclass(frozen=True, order=True)
class DirectoryData(ProtocolDataclass):
    name: str = field(metadata={'type': string})
    files: List[FileData] = field(metadata={'type': array, 'subtype': FileData})

    @classmethod
    def deserialize(cls, pos: int, message: bytes):
        pos, name = string.deserialize(pos, message)
        pos, files = array.deserialize(pos, message, element_type=FileData)
        return pos, cls(name, files)

    def serialize(self) -> bytes:
        return string(self.name).serialize() + array(self.files).serialize(FileData)


def has_unparsed_bytes(pos: int, message: bytes) -> bool:
    return len(message[pos:]) > 0


def calc_md5(value: str) -> str:
    return hashlib.md5(value.encode('utf-8')).hexdigest()



import os
from typing import Dict, List, Optional, Set

from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings

from .network.network import ListeningConnectionErrorMode
from .shares.model import DirectoryShareMode


class SharedDirectorySettingEntry(BaseModel, validate_assignment=True):
    path: str
    share_mode: DirectoryShareMode = DirectoryShareMode.EVERYONE
    users: List[str] = Field(default_factory=list)


class WishlistSettingEntry(BaseModel, validate_assignment=True):
    query: str
    enabled: bool = True


class UpnpSettings(BaseModel, validate_assignment=True):
    enabled: bool = True
    lease_duration: int = 0


class ReconnectSettings(BaseModel, validate_assignment=True):
    auto: bool = False
    timeout: int = 10


class ServerSettings(BaseModel, validate_assignment=True):
    hostname: str = 'server.slsknet.org'
    port: int = 2416
    reconnect: ReconnectSettings = ReconnectSettings()


class ListeningSettings(BaseModel, validate_assignment=True):
    error_mode: ListeningConnectionErrorMode = ListeningConnectionErrorMode.CLEAR
    port: int = 60000
    obfuscated_port: int = 60001


class PeerSettings(BaseModel, validate_assignment=True):
    obfuscate: bool = False


class NetworkLimitSettings(BaseModel, validate_assignment=True):
    upload_speed_kbps: int = 0
    download_speed_kbps: int = 0


class NetworkSettings(BaseModel, validate_assignment=True):
    server: ServerSettings = Field(default_factory=ServerSettings)
    listening: ListeningSettings = Field(default_factory=ListeningSettings)
    peer: PeerSettings = Field(default_factory=PeerSettings)
    upnp: UpnpSettings = Field(default_factory=UpnpSettings)
    limits: NetworkLimitSettings = Field(default_factory=NetworkLimitSettings)


class UserInfoSettings(BaseModel, validate_assignment=True):
    description: Optional[str] = None
    picture: Optional[bytes] = None


class CredentialsSettings(BaseModel, validate_assignment=True):
    username: str
    password: str
    info: UserInfoSettings = Field(default_factory=UserInfoSettings)

    def are_configured(self) -> bool:
        """Returns whether the credentials are correctly configured"""
        return self.username and self.password is not None

class SearchSettings(BaseModel, validate_assignment=True):
    wishlist: List[WishlistSettingEntry] = Field(default_factory=list)


class TransferLimitSettings(BaseModel, validate_assignment=True):
    upload_slots: int = 2


class TransfersSettings(BaseModel, validate_assignment=True):
    limits: TransferLimitSettings = Field(default_factory=TransferLimitSettings)
    report_interval: float = 0.250


class SharesSettings(BaseModel, validate_assignment=True):
    scan_on_start: bool = True
    download: str = os.getcwd()
    directories: List[SharedDirectorySettingEntry] = Field(default_factory=list)


class RoomsSettings(BaseModel, validate_assignment=True):
    auto_join: bool = True
    private_room_invites: bool = True
    favorites: Set[str] = Field(default_factory=set)


class UsersSettings(BaseModel, validate_assignment=True):
    friends: Set[str] = Field(default_factory=set)
    blocked: Set[str] = Field(default_factory=set)


class InterestsSettings(BaseModel, validate_assignment=True):
    liked: Set[str] = Field(default_factory=set)
    hated: Set[str] = Field(default_factory=set)


class DebugSettings(BaseModel, validate_assignment=True):
    search_for_parent: bool = True
    ip_overrides: Dict[str, str] = Field(default_factory=dict)
    log_connection_count: bool = False


class Settings(BaseSettings, validate_assignment=True):
    network: NetworkSettings = Field(default_factory=NetworkSettings)
    credentials: CredentialsSettings = Field(default_factory=CredentialsSettings)  # type: ignore
    searches: SearchSettings = Field(default_factory=SearchSettings)
    shares: SharesSettings = Field(default_factory=SharesSettings)
    users: UsersSettings = Field(default_factory=UsersSettings)
    rooms: RoomsSettings = Field(default_factory=RoomsSettings)
    interests: InterestsSettings = Field(default_factory=InterestsSettings)
    transfers: TransfersSettings = Field(default_factory=TransfersSettings)
    debug: DebugSettings = Field(default_factory=DebugSettings)



import logging

from .base_manager import BaseManager
from .events import (
    on_message,
    build_message_map,
    EventBus,
    MessageReceivedEvent,
    UserDirectoryEvent,
    UserSharesReplyEvent,
)
from .protocol.messages import (
    PeerDirectoryContentsRequest,
    PeerDirectoryContentsReply,
    PeerSharesRequest,
    PeerSharesReply,
    PeerUserInfoReply,
    PeerUserInfoRequest,
)
from .network.connection import PeerConnection
from .network.network import Network
from .settings import Settings
from .shares.manager import SharesManager
from .transfer.interface import UploadInfoProvider
from .user.manager import UserManager
from .utils import ticket_generator


logger = logging.getLogger(__name__)


class PeerManager(BaseManager):
    """Peer manager is responsible for handling peer messages"""

    def __init__(
            self, settings: Settings,
            event_bus: EventBus,
            user_manager: UserManager,
            shares_manager: SharesManager, upload_info_provider: UploadInfoProvider,
            network: Network):
        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._network: Network = network
        self._user_manager: UserManager = user_manager
        self._shares_manager: SharesManager = shares_manager
        self._upload_info_provider: UploadInfoProvider = upload_info_provider

        self._ticket_generator = ticket_generator()

        self._MESSAGE_MAP = build_message_map(self)

        self.register_listeners()

    def register_listeners(self):
        self._event_bus.register(
            MessageReceivedEvent, self._on_message_received)

    # Peer messages

    @on_message(PeerSharesRequest.Request)
    async def _on_peer_shares_request(self, message: PeerSharesRequest.Request, connection: PeerConnection):
        if not connection.username:
            logger.warning(
                "got PeerSharesRequest for a connection that wasn't properly initialized")
            return

        visible, locked = self._shares_manager.create_shares_reply(connection.username)
        await connection.send_message(
            PeerSharesReply.Request(
                directories=visible,
                locked_directories=locked
            )
        )

    @on_message(PeerSharesReply.Request)
    async def _on_peer_shares_reply(self, message: PeerSharesReply.Request, connection: PeerConnection):
        if not connection.username:
            logger.warning(
                "got PeerSharesRequest for a connection that wasn't properly initialized")
            return

        logger.info(f"PeerSharesReply : from username {connection.username}, got {len(message.directories)} directories")

        user = self._user_manager.get_user_object(connection.username)

        await self._event_bus.emit(
            UserSharesReplyEvent(
                user=user,
                directories=message.directories,
                locked_directories=message.locked_directories or [],
                raw_message=message
            )
        )

    @on_message(PeerDirectoryContentsRequest.Request)
    async def _on_peer_directory_contents_req(self, message: PeerDirectoryContentsRequest.Request, connection: PeerConnection):
        if not connection.username:
            logger.warning(
                "got PeerDirectoryContentsRequest for a connection that wasn't properly initialized")
            return

        directories = self._shares_manager.create_directory_reply(message.directory)
        await connection.send_message(
            PeerDirectoryContentsReply.Request(
                ticket=message.ticket,
                directory=message.directory,
                directories=directories
            )
        )

    @on_message(PeerDirectoryContentsReply.Request)
    async def _on_peer_directory_contents_reply(self, message: PeerDirectoryContentsReply.Request, connection: PeerConnection):
        if not connection.username:
            logger.warning(
                "got PeerDirectoryContentsReply for a connection that wasn't properly initialized")
            return

        user = self._user_manager.get_user_object(connection.username)
        await self._event_bus.emit(
            UserDirectoryEvent(
                user=user,
                directory=message.directory,
                directories=message.directories,
                raw_message=message
            )
        )

    @on_message(PeerUserInfoRequest.Request)
    async def _on_peer_user_info_request(self, message: PeerUserInfoRequest.Request, connection: PeerConnection):
        if not connection.username:
            logger.warning(
                "got PeerSharesRequest for a connection that wasn't properly initialized")
            return

        description = self._settings.credentials.info.description or ""
        picture = self._settings.credentials.info.picture

        await connection.send_message(
            PeerUserInfoReply.Request(
                description=description,
                has_picture=bool(picture),
                picture=picture,
                upload_slots=self._upload_info_provider.get_upload_slots(),
                queue_size=self._upload_info_provider.get_queue_size(),
                has_slots_free=self._upload_info_provider.has_slots_free()
            )
        )

    async def _on_message_received(self, event: MessageReceivedEvent):
        message = event.message
        if message.__class__ in self._MESSAGE_MAP:
            await self._MESSAGE_MAP[message.__class__](message, event.connection)



from dataclasses import dataclass, field
import datetime
from enum import auto, Enum
import re
from typing import List, Optional

from ..protocol.primitives import FileData


class SearchType(Enum):
    NETWORK = auto()
    USER = auto()
    ROOM = auto()
    WISHLIST = auto()


@dataclass
class ReceivedSearch:
    """Used for keeping track of searches received from the distributed parent
    or server
    """
    username: str
    query: str
    result_count: int


@dataclass
class SearchQuery:
    query: str
    include_terms: List[str] = field(default_factory=list)
    exclude_terms: List[str] = field(default_factory=list)
    wildcard_terms: List[str] = field(default_factory=list)

    @classmethod
    def parse(cls, query: str) -> 'SearchQuery':
        """Parses the given query string into terms and creates a new
        `SearchQuery` object
        """
        obj = cls(query)

        terms = query.split()
        for term in terms:
            # Ignore terms containing only non-word chars
            l_term = term.lower()
            if not re.search(r'[^\W_]', l_term):
                continue

            if term.startswith('*'):
                obj.wildcard_terms.append(l_term[1:])
            elif term.startswith('-'):
                obj.exclude_terms.append(l_term[1:])
            else:
                obj.include_terms.append(l_term)

        return obj

    def has_inclusion_terms(self) -> bool:
        """Return whether this query has any valid inclusion terms"""
        return bool(self.include_terms) or bool(self.wildcard_terms)


@dataclass
class SearchResult:
    """Search result received from a user"""
    ticket: int
    username: str

    has_free_slots: bool = False
    avg_speed: int = 0
    queue_size: int = 0

    shared_items: List[FileData] = field(default_factory=list)
    locked_results: List[FileData] = field(default_factory=list)


@dataclass
class SearchRequest:
    """Search request we have made"""
    ticket: int
    query: str
    search_type: SearchType = SearchType.NETWORK

    room: Optional[str] = None
    username: Optional[str] = None
    results: List[SearchResult] = field(default_factory=list)
    started: datetime.datetime = field(default_factory=datetime.datetime.now)
